#!/usr/bin/env python3
#Author: Yutong Wang at Samueli School of ENgineering, UCI


#------inclusions------
import os
from os import walk
import string
import sys
from collections import defaultdict

#------define variables------
#arguments
detail = False
debug = False
graph = False
u_help = False
u_out = False
u_in = False
#variables for script
has_main = False
outname = "Main_flat.cpp"
filepath = []

#the path of the script should probably change to make this work in the build
path = os.getcwd()

path2main = ""
path2folder = ""
main_name = ""
suffix = ""

args = sys.argv
arg_count = len(sys.argv)-1

#classes for graph and file hierarchy
class c_file:
	def __init__(self,fname=None,level=None):
		self.fname = fname
		self.level = level
		self.parent = None
		self.nextval = None
		self.preval = None
		self.include = []

class Graph:
	c_files = []
	h_files = []
	include = set()
	visited_g = set()
	printorder = []
	def __init__(self):
		self.head = None
		self.end = None
		self.graph = defaultdict(list)
		self.maxlvl = 0

	def listprint(self):
		printval = self.head
		while printval is not None:
			print(printval.fname +"\tlevel " +str(printval.level)+"\t\tParent: "+str(printval.parent))
			printval = printval.nextval

	def addEnd(self,newdata,newdata2,parent):
		newFile = c_file(newdata,newdata2)
		if self.head is None:
			self.head = newFile
			self.end = newFile
			return
		last = self.end
		newFile.parent = parent
		newFile.preval = last
		last.nextval = newFile
		self.end = newFile
		if newdata2>self.maxlvl:
			self.maxlvl = newdata2

	def addEdge(self,u,v):
		self.graph[u].append(v)

	def DFS(self):
		visited = self.visited_g
		self.DFS_recur(main_name,visited,None,0)

	def find(self, name):
		val = self.head
		while val is not None:
			if val.fname ==	name:
				return val
			val = val.nextval
		return None

	def DFS_recur(self,current,visited,prev,level):
		visited.add(current)
		self.addEnd(current,level,prev)
		f = open(filepath[filepath.index(current)+1]+current,"r")
		for line in f:
			if (line.find("#include")==0):
				line = deal_string(line)
				if filepath.count(line)!=0:
					self.addEdge(current,line)
					self.end.include.append(line)
		if debug:
			print(current)
			print(self.end.include)
		for neighbor in self.graph[current]:
			if neighbor not in visited:
				self.DFS_recur(neighbor,visited,current,level+1)
	def draw(self):
		val = main_name
		self.drawGraph(val,0)
		#self.printorder.reverse()
		if debug:
			print("PRINT ORDER:",self.printorder)	

	def drawGraph(self, val, lvl):
		printval = self.find(val)
		if graph:
			print("\t"*lvl+"|--------"+printval.fname)
		self.printorder.append(printval.fname)
		for item in printval.include:
			self.drawGraph(item,lvl+1)
		return

	def writefile(self):
		printed = []
		if not graph:
			self.draw()
		self.printorder.reverse()
		visited = self.visited_g
		c_files = self.c_files
		h_files = self.h_files
		inclusion = []
		f_w = open(path2folder+"temp","w")
		f_wf = open(path2folder+outname,"w")
		
		for item in self.printorder:
			if item not in printed:
				fname = item
				fname_c = fname.split(".")
				fname = fname_c[0]
				for fs in filepath:
					aa = fs.split(".")
					a = aa[0];
					if(fname == a)and(suffix in fs)and(fs not in visited):
						c_files.append(fs)

#					if(fs.find(fname)!=-1)and(suffix in fs)and(fs not in visited):
#						c_files.append(fs)

				if item is not main_name:
					f_r = open(filepath[filepath.index(item)+1]+item,"r")
					write_f(f_r,f_w,self.include)
					h_files.append(item)
					f_r.close()
				printed.append(item)
		
		c_files.append(main_name)
		for fs in c_files:
			f_r = open(filepath[filepath.index(fs)+1]+fs,"r")
			write_f(f_r,f_w,self.include)
			f_r.close()

		f_w.close()
		f_w = open(path2folder+"temp","r")
		for line in self.include:
			f_wf.write(line)
		for line in f_w:
			f_wf.write(line)
		f_w.close()
		f_wf.close()
		os.remove(path2folder+"temp")
		if debug or detail:
			print("SYSTEM_INCLUSION: "+str(self.include))
			print("H_FILES: "+str(h_files))
			print("C_FILES: "+str(c_files))
		

#------define string function------
def write_f(f_r,f_w,include):
	for line in f_r:
		if("#include" not in line):
			f_w.write(line)
		line_2 = deal_string(line)
		if("#include" in line) and (line_2 not in filepath):
			include.add(line)
	f_w.write("\n");
			

def deal_string(s):

	s = s.replace(" ","")
	s = s.replace("#include","")
	s = s.replace("\n","")
	s = s.replace("\"","")
	if "/" in s:
		a = s.split("/")
		s = a[1]
	return s

def simple_alg(main_name):
	print("test")

#------take arguments------
while(arg_count > 0):
	if (args.count("-h")!=0) or (args.count("-help")!=0):
		print("""
User Help:
use -v argument to show included files
use -g argument to show graph(tree) of file dependency
use -x argument to auto compile the flattened file using RISC
use -i #name# argument to indicate the main file, no need to input multiple files
use -o #name# argument to change the name of output file, if not default to Main_flat.cpp
use -d argument to enter debug mode
		""")
		quit()
	if (args.count("-d")!=0):
		debug = True
		print("------DEBUG MODE ON------")
		print("REVISION 7")
		arg_count -= 1
	if (args.count("-v")!=0):
		detail = True
		arg_count -= 1 
	if (args.count("-g")!=0):
		graph = True
		arg_count -= 1
	if (args.count("-i")!=0):
		u_in = True
		main_name = args[args.index("-i")+1]
		if debug:
			print("user input main name: "+ main_name)
		arg_count -= 2
	if (args.count("-o")!=0):
		u_out = True
		outname = args[args.index("-o")+1]
		if debug:
			print("user output name:" + outname)
		arg_count -= 2
	if arg_count > 0:
		print(arg_count)
		print("Found unrecognized arguments, exiting...")
		print("Use -h or -help for help with the script")
		quit()
	
#------script starts------
#Get a list of paths and file names, find main based on auto search or user input
for root,dirs,files in os.walk(path,topdown=False):
	for name in files:
		filepath.append(name)
		filepath.append(root+"/")
		if has_main and not u_in:
			print("----Found multiple files containting name main or Main, exiting....Please use manual input----\n")
			quit()
		if u_in:
			if main_name == name:
				if debug:
					print("FOUND USER INPUT!")
				path2main = os.path.join(root,name)
				path2folder = root + "/"
				has_main = True
		elif (("main" in name) or ("Main" in name))and not u_in:
			path2main = os.path.join(root,name)
			main_name = name
			path2folder = root + "/"
			has_main = True

if has_main and not u_in:
	print("Found main file, the path is: " + path2main)
elif has_main and u_in:
	print("User input name for main, the path is: " + path2main)
else:
	print("ERROR: input file not found, exiting....")
	quit()

suffix_temp = main_name.split(".")
suffix = "."+ suffix_temp[1]
if debug:
#	print(filepath)
	print("HAS_MAIN = "+str(has_main))
	print("PATH2MAIN = "+path2main)
	print("PATH2FOLDER = "+path2folder)
	print("------PATH_END------\n")

#exploring the files starting from mian
g = Graph()
g.DFS()
#GRAPH
if debug:
	print("\n")
	g.listprint()
if graph:
	print("---------GRAPH----------")
	print("NOTICE: ALL the C files are still printed in the correct order though they are not shown in the graph\n")
	g.draw()

g.writefile()
print("---DONE FLATTENING---")

