#ifndef MODULE_INSTANCE_H_INCLUDED_
#define MODULE_INSTANCE_H_INCLUDED_

#include <string>
#include <vector>
#include <map>

#include "rose.h"

#include "instance.h"
#include "port.h"
#include "object.h"

#include "class.h" // included for get_definition
#include "module.h" // included for get_definition

/*TLM 2.0*/
#include "socket.h"

namespace risc {

/** \class ModuleInstance
 *  \brief This class represents a instance of a module
 *
 *  A module instance has a type pointer and rose ast pointer.
 */
class ModuleInstance: public Instance {

public:

  /**
   * \brief default constructor
   */
  explicit ModuleInstance(SgVariableDefinition *ast_node, Module *ir_def);

  /**
   * \brief This constructor is used for classes which derived from sc_module
   *        Currently, this is only sc_channel
   */
  explicit ModuleInstance(SgVariableDefinition *ast_node, Type type);

  /**
   * \fn       get_definition
   * \brief    returns this associated definition of the module instance
   */
  virtual Module* get_definition();

  /**
   * \fn    add_mapping(Port *port, Object *object)
   * \brief This is a helper functio to add a mapping
   */
  void add_mapping(Port *port, Object *object);

  /**
   * \fn    add_socket_mapping(Socket *socket_from, Socket *socket_to)
   * \brief This is a helper functio to add a mapping
   */
  void add_socket_mapping(Socket *socket_from, Socket *socket_to);

  /**
   * \fn    add_backward_socket_mapping(Socket *socket_from, Socket *socket_to)
   * \brief This is a helper functio to add a mapping
   */
  void add_backward_socket_mapping(Socket *socket_from, Socket *socket_to);

  /**
   * \fn    get_instance_qualified_name
   * \brief return the qualified name of the module instance
   *        for example: ::top::prod_inst
   */
  std::string get_instance_qualified_name();
  
  std::map<Port*, Object*> mapping_;

  /* TLM 2.0 */
  std::map<Socket*, Socket*> socket_mapping_;
  std::map<Socket*, Socket*> backward_socket_mapping_;
private:

  /**
   * \brief This pointer points to the corresponding definition in the
   *        internal representation
   */
  Module* module_definition_;

  /**
   * \brief Disabled copy constructor
   */
  ModuleInstance(const ModuleInstance &md);
};

typedef std::vector<ModuleInstance*>
            ModuleInstanceVector;
typedef std::vector<ModuleInstance*>::iterator
            ModuleInstanceVectorIter;
typedef std::vector<ModuleInstance*>::const_iterator
            ModuleInstanceVectorConstIter;

}; // end namespace risc

#endif /* MODULE_INSTANCE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
