#include "combined_data_conflict_table.h"

#include<iostream>
#include<cstring>


#include "graph_helper.h"

#include "../internal_representation/cthread.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/hierarchical_channel_instance.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/method.h"
#include "../internal_representation/module.h"
#include "../internal_representation/module_instance.h"
#include "../internal_representation/thread.h"
#include "../internal_representation/path_instance_mapper.h"
#include "../internal_representation/static_path_instance_mapper.h"
#include "../internal_representation/dynamic_path_instance_mapper.h"

risc::sg::CombinedDataConflictTable::CombinedDataConflictTable(
    risc::sg::DataConflictTable &dct,
    PathInstanceMapper *path_instance_mapper):
  dct_(dct),
  path_instance_mapper_(path_instance_mapper),
  prediction_steps_(0)
{
  number_of_segments_ = boost::num_vertices(dct_.segment_graph_.graph_);
  max_instances_ = path_instance_mapper_->max_instances_;

  instance_id_to_index_lookup_ = new int*[max_instances_ + 1];

  // We are adding one additional row to safe the segment id
  for(int i=0; i < max_instances_ + 1; i++) {

    instance_id_to_index_lookup_[i] = new int[number_of_segments_];
  }

  // Here, the first row of the instance_id_to_index_lookup table is filled
  // with the segment ids. We have to do that because the segment ids do not
  // incease by one.
  std::pair<VertexIterator, VertexIterator> vp;
  int i = 0;
  for(vp = boost::vertices(dct_.segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    instance_id_to_index_lookup_[0][i]
      = dct_.segment_graph_.graph_[*(vp.first)].id_;
  }

  // We are filling all array fields with -1 to indicate that a specific field
  // is not used. We need that information for the regression tests.
  for(int i = 1; i < max_instances_ + 1; i++) {
    for(int j=0; j < number_of_segments_; j++) {
      instance_id_to_index_lookup_[i][j] = -1;
    }
  }

  // Here, we are filling the instance_id_to_index_lookup_ table. Each field
  // in this table needs an unique value. The value is the associated row/column
  // in the conflict table.
  int index = 0;

  // We are iterating over all the instances in the design
  for(std::vector<PathInstanceMapper::PathWithID>::iterator
      iter  = path_instance_mapper_->path_to_id_mapping_.begin();
      iter != path_instance_mapper_->path_to_id_mapping_.end();
      iter++) {

    // Adding the value of 1 because the first row is reserved for the
    // segment id
    int row = (*iter).second + 1;

    Object *object = (*iter).first.get_object();

    ModuleInstance *module_instance = dynamic_cast<ModuleInstance*>(object);

    Module *module = NULL;

    if(module_instance) {
      module = module_instance->get_definition();
    }

    if(!module) {

      module = dynamic_cast<Module*>((*iter).first.path_.back().first);
    }

    if(module) {

      for(ThreadVectorConstIter
          thread_iter  = module->threads_.begin();
          thread_iter != module->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            dct_.segment_graph_, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorConstIter
          cthread_iter  = module->cthreads_.begin();
          cthread_iter != module->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            dct_.segment_graph_, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorConstIter
          method_iter  = module->methods_.begin();
          method_iter != module->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            dct_.segment_graph_, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }

    // This is the same like for the modules. So, please see the documents above
    HierarchicalChannelInstance *hierarchical_channel_instance
      = dynamic_cast<HierarchicalChannelInstance*>(object);

    if(hierarchical_channel_instance) {

      HierarchicalChannel *hierarchical_channel
        = hierarchical_channel_instance->get_definition();

      for(ThreadVectorConstIter
          thread_iter  = hierarchical_channel->threads_.begin();
          thread_iter != hierarchical_channel->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            dct_.segment_graph_, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorConstIter
          cthread_iter  = hierarchical_channel->cthreads_.begin();
          cthread_iter != hierarchical_channel->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            dct_.segment_graph_, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorConstIter
          method_iter  = hierarchical_channel->methods_.begin();
          method_iter != hierarchical_channel->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            dct_.segment_graph_, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }
  }

  size_of_conflict_table_ = index;

  // Allocating the memory for the VertexDescriptor array
  vertex_lookup_ = new VertexDescriptor[number_of_segments_];

  // Here, we are filling the vertex lookup. The array field at position is the
  // segment with the index instance_id_to_index_lookup_[0][i]
  for(vp = boost::vertices(dct_.segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    int i = 0;
    for(; i < number_of_segments_; i++) {
      if(instance_id_to_index_lookup_[0][i]
          == dct_.segment_graph_.graph_[*(vp.first)].id_) {

        vertex_lookup_[i] = *(vp.first);
        break;
      }
    }
  }

  // creating the conflict table
  combined_table_ = new int[size_of_conflict_table_ * size_of_conflict_table_];

  for(int i=0;i<size_of_conflict_table_;i++) {

    for(int j=0;j<size_of_conflict_table_;j++) {
      combined_table_[i * size_of_conflict_table_ + j] = 0;
    }
  }

  compute_combined_table();
}

risc::sg::CombinedDataConflictTable::~CombinedDataConflictTable()
{
  delete[] combined_table_;
}

void risc::sg::CombinedDataConflictTable::print_conflict_table()
{
  for(int i=0; i < size_of_conflict_table_; i++) {
    for(int j=0; j < size_of_conflict_table_; j++) {

      std::cout << std::setw(4)
                << combined_table_[i * size_of_conflict_table_ + j] << " ";
    }

    std::cout << std::endl;
  }
  std::cout << std::endl;
}

void
risc::sg::CombinedDataConflictTable::print_conflict_table(std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i = 1; i < max_instances_ + 1; i++) {

    if(i == 1)
      myfile << "    <th>(Seg,Inst)</th>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {


      if(instance_id_to_index_lookup_[i][j] == -1) {
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <th>"
             << "(" << ids.first << "," << ids.second << ")" << "</th>"
             << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "  <tr>" << std::endl;

      if(instance_id_to_index_lookup_[i][j] == -1) {
        myfile << "  </tr>" << std::endl;
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <td>"
             << "(" << ids.first << "," << ids.second << ")" << "</td>"
             << std::endl;

      for(int k = 1; k < max_instances_ + 1; k++) {

        for(int l=0; l < number_of_segments_; l++) {

          if(instance_id_to_index_lookup_[k][l] == -1) {
            continue;
          }

          int row = instance_id_to_index_lookup_[i][j];
          int column = instance_id_to_index_lookup_[k][l];

          myfile << "    <td> "
                 << combined_table_[row * size_of_conflict_table_ + column]
                 << " </td>" << std::endl;
        }
      }

      myfile << "  </tr>" << std::endl;
    }
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

void
risc::sg::CombinedDataConflictTable::print_lookup_table(std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i=0; i < number_of_segments_ + 1; i++) {

    if(i == 0) {
      myfile << "    <th> </th>" << std::endl;

    } else {
      myfile << "    <th>"
             << instance_id_to_index_lookup_[0][i-1]
             << "</th>" << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i=1; i < max_instances_ + 1; i++) {

    myfile << "  <tr>" << std::endl;
    myfile << "<td>" << i-1 << "</td>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "<td>" << instance_id_to_index_lookup_[i][j]
             << "</td>" << std::endl;
    }
    myfile << "  </tr>" << std::endl;
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

int
risc::sg::CombinedDataConflictTable::segment_and_instance_id_to_index(
    int segment_id, int instance_id)
{
  int column_index = -1;
  for(int i = 0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id) {

      column_index = i;
    }
  }

  if(column_index == -1) {
    assert("segment_and_instance_id_to_index: This segment id does not exist"
        && 0);
  }

  return instance_id_to_index_lookup_ [instance_id + 1][column_index];
}

std::pair<int, int>
risc::sg::CombinedDataConflictTable::index_to_segment_and_instance_id(int index)
{
  std::pair<int, int> result;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      if(instance_id_to_index_lookup_[i][j] == index) {

        result.first = instance_id_to_index_lookup_[0][j];
        result.second = i - 1;
        return result;
      }
    }
  }

  std::cout << "index: " << index << std::endl;
  assert("index_to_segment_and_instance_id: Index does not exist" && 0);
}

int risc::sg::CombinedDataConflictTable::get_max_instances(int segment_id)
{
  for(int i=0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id)  {

      int counter = 1;
      for(; counter < max_instances_ + 1; counter++) {
        if(instance_id_to_index_lookup_[counter][i] == -1) {
          break;
        }
      }

      return counter -1;
    }
  }

  assert("Cannot determine max instances" && 0);
  return -1;
}

void
risc::sg::CombinedDataConflictTable::print_lookup_table()
{

  for(int i=0; i < max_instances_ + 1; i++) {

    if(i == 0) {

      std::cout << "    ";
    } else {
      std::cout << std::setw(4) << i-1;
    }

    for(int j=0; j < boost::num_vertices(dct_.segment_graph_.graph_); j++) {

      std::cout << std::setw(4) << instance_id_to_index_lookup_[i][j] << " ";
    }

    std::cout << std::endl;
  }
}

int risc::sg::CombinedDataConflictTable::max_prediction_steps()
{
  return prediction_steps_;
}

void risc::sg::CombinedDataConflictTable::print_conflict_table(bool *table)
{
  for(int i=0; i < size_of_conflict_table_; i++) {
    for(int j=0; j < size_of_conflict_table_; j++) {

      std::cout << std::setw(4)
                << table[i * size_of_conflict_table_ + j] << " ";
    }

    std::cout << std::endl;
  }
  std::cout << std::endl;
}

void risc::sg::CombinedDataConflictTable::compute_combined_table()
{
  int size = size_of_conflict_table_;
  // For the overall strategy please refer to the DATE 13 paper
  // 'Optimized Out-of-Order Parallel Discrete Event Simulation Using
  // Predictions"
  //
  // This implementation is a brute force approach and it is not optimized
  // for run time concerns.

  // Here, we prepare the adjacent matrix
  // First, we set the diagonal elements to 1 and the rest to 0
  bool *mod_adj_matrix = new bool[size * size];

  for(int i=0; i < size_of_conflict_table_; i++) {
    for(int j=0; j < size_of_conflict_table_; j++) {

      if(i == j) {
        mod_adj_matrix[i * size + j] = true;
      } else {
        mod_adj_matrix[i * size + j] = false;
      }
    }
  }

  // Second, we set all elements to 1 where the segments have a transition.
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(dct_.segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;

    boost::tie(out_edge_begin, out_edge_end)
      = boost::out_edges(*(vp.first), dct_.segment_graph_.graph_);

    if(out_edge_begin != out_edge_end) {

      for(out_edge_iter  = out_edge_begin;
          out_edge_iter != out_edge_end;
          out_edge_iter++) {

        int from_sg_id = dct_.segment_graph_.graph_[*(vp.first)].id_;
        int to_sg_id = dct_.segment_graph_.graph_[
             boost::target( *out_edge_iter, dct_.segment_graph_.graph_) ].id_;

        for(int instance_from = 0;
            instance_from < get_max_instances(from_sg_id);
            instance_from++) {

          int from_index = segment_and_instance_id_to_index(
                            from_sg_id, instance_from);

          // TODO I don't know how to deal with this one
          //for(int instance_to = 0;
          //    instance_to < get_max_instances(to_sg_id);
          //    instance_to++) {

          //  int to_index = segment_and_instance_id_to_index(
          //                  //to_sg_id, instance_to);
          //                  to_sg_id, instance_from);

          //  mod_adj_matrix[from_index * size + to_index] = 1;
          //}

          int to_index = segment_and_instance_id_to_index(
                          to_sg_id, instance_from);

          mod_adj_matrix[from_index * size + to_index] = true;
        }
      }
    }
  }

  // Here we are computing the prediction steps.
  // the general equation is
  //  CT_0 = CT
  //  CT_n = A * CT_(n-1)

  // These are helper vectors
  bool *ctn = new bool[size * size];
  bool *ctn_p1 = new bool[size * size];

  for(int i=0; i<size; i++) {

    for(int j=0; j<size; j++) {
      if(dct_.conflict_table_[i][j]) {

        ctn[i * size + j] = 1;
        combined_table_[i * size + j] = 1;

      } else {
        ctn[i * size + j] = 0;
        combined_table_[i * size + j] = 0;
      }
    }
  }

  // initializing ctn_p1 with ctn
  std::memcpy(ctn_p1, ctn, size * size * sizeof(bool));

  int m = 1;
  while(true) {

    // ctnp1 = mod_adj_matrix * ctn
    for(int i = 0; i < size; i++) {
      for(int j = 0; j < size; j++) {

        // We can skip this iteration because the flag is already set
        if(ctn_p1[i * size + j])
          continue;

        for (int k = 0; k < size; k++) {
          ctn_p1[i * size + j]
            |= mod_adj_matrix[i * size + k] && ctn[k * size + j];
        }
      }
    }

    // update the combined_table_
    bool has_changed = false;

    for(int i = 0; i < size; i++) {
      for(int j = 0; j < size; j++) {
        if(ctn[i * size + j] != ctn_p1[i * size + j]) {
          has_changed = true;
          combined_table_[i * size + j] = m+1;
        }
      }
    }

    if(!has_changed) {
      break;
    }

    // ctn = ctn_p1
    std::memcpy(ctn, ctn_p1, size * size * sizeof(bool));

    m++;
  }

  prediction_steps_ = m + 1;

  delete[] ctn;
  delete[] ctn_p1;
  delete[] mod_adj_matrix;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
