#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"
#include <string>
#include "systemc.h"
using namespace sc_core;
using namespace std;


/*------------------initiator---------------------*/
class lt_initiator :  public sc_core::sc_module 
{
public:
    lt_initiator( sc_core::sc_module_name name);
  	void initiator_thread (void);
    SC_HAS_PROCESS(lt_initiator);
   
    tlm_utils::simple_initiator_socket<lt_initiator> initiator_socket;
 	int init_data;
}; 

lt_initiator::lt_initiator( sc_module_name name )
: sc_module           (name)                      // initialize module name
, initiator_socket    ("initiator_socket")        // initiator socket
{                
	init_data = 42;
	SC_THREAD(initiator_thread);                  
}

void lt_initiator::initiator_thread(void)   ///< initiator thread
{  
	tlm::tlm_generic_payload trans_init; 
	trans_init.set_data_ptr( reinterpret_cast<unsigned char*>(&init_data));
	trans_init.set_data_length(4);
	sc_time delay_init = SC_ZERO_TIME;       

	initiator_socket->b_transport(trans_init, delay_init);
} 

/*------------------target---------------------*/
class lt_target : public sc_core::sc_module           	
{
public:	  
 
	lt_target( sc_core::sc_module_name   module_name);

	void custom_b_transport( 
		tlm::tlm_generic_payload  &trans,
		sc_core::sc_time          &delay_time
  	);

	SC_HAS_PROCESS(lt_target);
	tlm_utils::simple_target_socket<lt_target>  target_socket;

};

lt_target::lt_target( sc_core::sc_module_name module_name )
: sc_module (module_name)
, target_socket ("target_socket")
{
	target_socket.register_b_transport(this, &lt_target::custom_b_transport);
}

void                                        
lt_target::custom_b_transport( 
	tlm::tlm_generic_payload  &trans,
	sc_core::sc_time          &delay_time          
)
{
	wait(2,SC_NS);
	unsigned char* ptr = trans.get_data_ptr();
	int len = trans.get_data_length();
	int d;
	memcpy(&d, ptr, len);

	wait(1,SC_NS);
	std::cout << "d = " << d << std::endl;
}

/*------------------top---------------------*/

class lt_top : public sc_core::sc_module                   
{
public:
		
	lt_top ( sc_core::sc_module_name name); 
	lt_target               m_target_1;          
	lt_initiator            m_initiator_1;  
};



lt_top::lt_top( sc_core::sc_module_name name )
: sc_core::sc_module(name)
, m_target_1("m_target_1")
, m_initiator_1("m_initiator_1")
{
  std::cout << "in top.cpp" << std::endl;
  m_initiator_1.initiator_socket(m_target_1.target_socket);
}

int sc_main(int   argc, char  *argv[] )
{
  lt_top top("top");                    // instantiate a exmaple top module

  sc_core::sc_start();                  // start the simulation

  return 0;                             // return okay status
}
