#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"

#include <string>
#include "systemc.h"
using namespace sc_core;
using namespace std;


/****************************************/

class lt_initiator :  public sc_core::sc_module 
{
public:
    lt_initiator( sc_core::sc_module_name name);
  	void initiator_thread (void);                    
    SC_HAS_PROCESS(lt_initiator);
    tlm_utils::simple_initiator_socket<lt_initiator> initiator_socket;
}; 

lt_initiator::lt_initiator( sc_module_name name )
: sc_module           (name)                      // initialize module name
, initiator_socket    ("initiator_socket")        // initiator socket
{                
	SC_THREAD(initiator_thread);                  
}

void lt_initiator::initiator_thread(void)   ///< initiator thread
{  
	tlm::tlm_generic_payload trans; 
	sc_time delay = SC_ZERO_TIME;       // Create delay objects 
	initiator_socket->b_transport(trans, delay);
} 

/****************************************/

class lt_target : public sc_core::sc_module           	
{
public:	  
 
	lt_target( sc_core::sc_module_name   module_name);

	void custom_b_transport( 
		tlm::tlm_generic_payload  &gp,
		sc_core::sc_time          &delay_time
  	);

	SC_HAS_PROCESS(lt_target);
	tlm_utils::simple_target_socket<lt_target>  m_memory_socket;

	std::string              m_target_memory;
};

lt_target::lt_target( sc_core::sc_module_name module_name )
: sc_module (module_name)
, m_memory_socket ("memory_socket_1")
, m_target_memory ("Hello World")
{
  m_memory_socket.register_b_transport(this, &lt_target::custom_b_transport);
}

void                                        
lt_target::custom_b_transport( 
	tlm::tlm_generic_payload  &gp,
	sc_core::sc_time          &delay_time             // delay time 
)
{
 
  int length = gp.get_data_length();
  wait(1,SC_NS);
  std::cout << "custom_b_transport called" << std::endl;
    
}

/****************************************/

class lt_wrapper : public sc_core::sc_module                   
{
public:
		
	lt_wrapper ( sc_core::sc_module_name name); 
	//SimpleBusLT<1, 1>       m_bus;                ///< simple bus
	lt_initiator            m_initiator_1;          ///< instance 1 initiator
	tlm::tlm_initiator_socket< > top_initiator_socket;
};

lt_wrapper::lt_wrapper( sc_core::sc_module_name name )
: sc_core::sc_module(name)
, m_initiator_1("m_initiator_1")
, top_initiator_socket ("top_initiator_socket")    
{
  std::cout << "in wrapper" << std::endl;
  m_initiator_1.initiator_socket(top_initiator_socket);
}


class lt_wrapper2 : public sc_core::sc_module                   
{
public:
		
	lt_wrapper2 ( sc_core::sc_module_name name); 
	//SimpleBusLT<1, 1>       m_bus;                ///< simple bus
	lt_target            m_target_1;          ///< instance 1 initiator
	tlm::tlm_target_socket< > top_target_socket;
};

lt_wrapper2::lt_wrapper2( sc_core::sc_module_name name )
: sc_core::sc_module(name)
, m_target_1("m_target_1")
, top_target_socket ("top_target_socket")    
{
  std::cout << "in wrapper2" << std::endl;
  this->top_target_socket(m_target_1.m_memory_socket);
}

/****************************************/

class lt_top : public sc_core::sc_module                   
{
public:
		
	lt_top ( sc_core::sc_module_name name); 
	//SimpleBusLT<1, 1>       m_bus;                ///< simple bus
	lt_wrapper            	m_wrapper_1;          ///< instance 1 initiator
	lt_wrapper2            	m_wrapper_2;          ///< instance 1 initiator
};

lt_top::lt_top( sc_core::sc_module_name name )
: sc_core::sc_module(name)
, m_wrapper_1("m_wrapper_1")
, m_wrapper_2("m_wrapper_2")
{
  std::cout << "in top" << std::endl;
  m_wrapper_1.top_initiator_socket(m_wrapper_2.top_target_socket);
}

int sc_main(int   argc, char  *argv[] )
{
  lt_top top("top");                    // instantiate a exmaple top module

  sc_core::sc_start();                  // start the simulation

  return 0;                             // return okay status
}
