// bug20160629.cc (discovered 6/29/16, put into regression tests 4/11/17, RD)

#include <stdio.h>
#include <assert.h>
#include "systemc.h"

#define ITERATIONS 10000

int t1counter = 0;
int t2counter = 0;

struct T1 : public sc_module
{
    T1(sc_module_name name_): sc_module(name_)
    {
	SC_HAS_PROCESS(T1);
	SC_THREAD(main);
    }

    void main(void)
    {
	int i;

	for(i=0; i<ITERATIONS; i++)
	{
	    wait_for_ack();
	    t1counter++;
	}
    }

    sc_event ack;

    void wait_for_ack(void)
    {
	ack.notify(SC_ZERO_TIME);
	wait(ack);
    }
};

struct T2 : public sc_module
{
    T2(sc_module_name name_): sc_module(name_)
    {
	SC_HAS_PROCESS(T2);
	SC_THREAD(main);
    }

    void main(void)
    {
	int i;

	for(i=0; i<ITERATIONS; i++)
	{
	    wait_for_ack();
	    t2counter++;
	}
    }

    sc_event ack;

    void wait_for_ack(void)
    {
	ack.notify(SC_ZERO_TIME);
	wait(ack);
    }
};

struct Top : public sc_module
{
    T1 t1;
    T2 t2;

    Top(sc_module_name name_): sc_module(name_),
	t1("t1"), t2("t2")
    { }
};

Top top("top");

int sc_main(int argc,char *argv[])
{
  printf("Starting... %d, %d\n", t1counter, t2counter);
  sc_start();
  printf("t1counter = %d\n", t1counter);
  printf("t2counter = %d\n", t2counter);
  assert(t1counter == ITERATIONS);
  assert(t2counter == ITERATIONS);
  printf("Done.\n");
  return 0;
}

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
