// bug20180302.cc:
// example "prodcons_rt_number" by A. Dingankar, 3/1/18
// modified by RD, 03/02/18
// added to regression test suite by RD, 04/17/18

// Needed for sc_fifo.trace()
#define DEBUG_SYSTEMC

#include <systemc.h> 

SC_MODULE(c_Cons) {
		public:
		sc_fifo_in<int> p_MsgQ;
		int MsgQ;
		int Loop_index;

		void init() {
			cout << sc_time_stamp() << ": /Cons/init: " << endl;
			Loop_index = 0; // RD
			}
		void consume() {
			cout << sc_time_stamp() << ": /Cons/consume: " << MsgQ << endl;
			}
		void behavior() {
			init(); 
			wait(10, SC_NS);
			while (1) {
				p_MsgQ.read(MsgQ);
				consume(); 
				wait(10, SC_NS);
				Loop_index++;
				}
		}

		SC_CTOR(c_Cons) {
		        SC_THREAD(behavior);
		}
};

SC_MODULE(c_Prod) {
		public:
		sc_fifo_out<int> p_MsgQ;
		int MsgQ;
		int Loop_index;

		void init() {
			cout << sc_time_stamp() << ": /Prod/init: " << endl;
			Loop_index = 0; // RD
			}
		void produce() {
			MsgQ = Loop_index; 
			cout << sc_time_stamp() << ": /Prod/produce: " << "MsgQ: " << MsgQ << endl;
			}
		void behavior() {
			init(); 
			wait(10, SC_NS);
			while (Loop_index < 4) {
				produce(); 
				wait(10, SC_NS);
				p_MsgQ.write(MsgQ);
				Loop_index++;
				}
		}

		SC_CTOR(c_Prod) {
		        SC_THREAD(behavior);
		}
};

class top : public sc_module {
	public:
		sc_fifo<int> inst_MsgQ;
		c_Cons inst_Cons;
		c_Prod inst_Prod;

	   top(sc_module_name name) : sc_module(name)
		, inst_MsgQ(2)
		, inst_Cons("Cons")
		, inst_Prod("Prod")
		{
		inst_Cons.p_MsgQ.bind(inst_MsgQ);
		inst_Prod.p_MsgQ.bind(inst_MsgQ);
    }
};

//sc_trace_file* Tf; // RD: RISC 0.4.0 work-around

int sc_main (int, char *[]) {
	top top1("Top1");
	sc_time TenNS(10, SC_NS);

	sc_trace_file* Tf;	// RD: RISC 0.4.0 bug, fixed in 0.4.1
	Tf = sc_create_vcd_trace_file("bug20180302_trace");
	Tf->set_time_unit(TenNS.to_seconds(), SC_SEC);

	top1.inst_MsgQ.trace(Tf);
//	sc_trace(Tf, top1.inst_MsgQ, "MsgQ"); // RD
	sc_trace(Tf, top1.inst_Cons.Loop_index, "Cons.Loop_index");
	sc_trace(Tf, top1.inst_Cons.MsgQ, "Cons.MsgQ"); // RD
	sc_trace(Tf, top1.inst_Prod.Loop_index, "Prod.Loop_index");
	sc_trace(Tf, top1.inst_Prod.MsgQ, "Prod.MsgQ"); // RD

	sc_start();
	sc_close_vcd_trace_file(Tf);
//	sc_stop(); // RD

	return 0;
}

