// bug20181203.cc:
// WAS: CannyA8_step3.cpp (ECPS203, Fall'18)

#include <stdio.h>
#include <assert.h>
#include "systemc.h"

#define IMG_NUM   30 /* number of images processed (1 or more) */
#define AVAIL_IMG 30 /* number of different image frames (1 or more) */

typedef unsigned char IMAGE;
typedef short int SIMAGE;
typedef float FKERNAL;
typedef float FIMAGE;

SC_MODULE(Stimulus){
	IMAGE imageout;
	sc_fifo_out<IMAGE> ImgOut;
	sc_fifo_out<sc_time> StartTime;

	void main(void)
	{
	   int i=0, n=0;
	   sc_time t;

	   for(i=0; i<IMG_NUM; i++)
	   {
	      n = i % AVAIL_IMG;
	      imageout = 42;
	      ImgOut.write(imageout);
	      t = sc_time_stamp();
	      printf("%9s: Stimulus sent frame%3d.\n", t.to_string().c_str(), n+1);
	      StartTime.write(t);
	   }
	}

	SC_CTOR(Stimulus)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(Monitor){
	IMAGE imagein;
	sc_fifo_in<IMAGE>  ImgIn;
	sc_fifo_in<sc_time> StartTime;

	void main(void)
	{
	   int i, n;
	   sc_time t1,t2,t,t3,t_delay;
	   sc_time ExpectedTotalTime(59320, SC_MS);
	   sc_time ExpectedLastDelay(15860, SC_MS);
	   char S1[20], ExpectedS1[20] = "1.820 seconds";
	   char S2[20], ExpectedS2[20] = "0.549 FPS";

	   for(i=0; i<IMG_NUM; i++)
	   {
	      ImgIn.read(imagein);
	      StartTime.read(t1);
		  t2=sc_time_stamp();
		  t=t2-t1;
		  t_delay = t2-t3;
		  t3=t2;
	      n = i % AVAIL_IMG;
	      printf("%9s: Monitor received frame%3d with%9s delay.\n",
			t2.to_string().c_str(), n+1, t.to_string().c_str());
	      printf("%9s: %5.3f seconds after previous frame, %5.3f FPS.\n",
			t2.to_string().c_str(),
			t_delay.to_double()/1000000000000,
			1000000000000/t_delay.to_double());
	      sprintf(S1, "%5.3f seconds", t_delay.to_double()/1000000000000);
	      sprintf(S2, "%5.3f FPS", 1000000000000/t_delay.to_double());
	   }
	   printf("%9s: Monitor exits simulation.\n", t2.to_string().c_str());
	   assert(t  == ExpectedLastDelay);
	   assert(t2 == ExpectedTotalTime);
	   assert(0 == strcmp(S1, ExpectedS1));
	   assert(0 == strcmp(S2, ExpectedS2));
	   sc_stop();	// done testing, quit the simulation
	}

	SC_CTOR(Monitor)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(DataIn)
{
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_out<IMAGE> ImgOut;
	IMAGE Image;

	void main()
	{
	   while(1)
	   {
	      ImgIn.read(Image);
	      ImgOut.write(Image);
	   }
	}
	
	SC_CTOR(DataIn)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(DataOut)
{
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_out<IMAGE> ImgOut;
	IMAGE Image;

	void main()
	{
	   while(1)
	   {
	      ImgIn.read(Image);
	      ImgOut.write(Image);
	   }
	}
	
	SC_CTOR(DataOut)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(Receive_Image)
{
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_out<IMAGE> ImgOut;
	IMAGE Image;

	void main()
	{
	   while(1)
	   {
	      ImgIn.read(Image);
	      ImgOut.write(Image);
	   }
	}
	
	SC_CTOR(Receive_Image)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(Gaussian_Kernel)
{
	sc_fifo_out<FKERNAL> G_Out1,G_Out2;
	sc_fifo_out<int> C_Out1,C_Out2;
	FKERNAL gaussian_kernel;
	int kernel_center;

	void main(void)
	{
	   int windowsize,       /* Dimension of the gaussian kernel. */
	       center;           /* Half of the windowsize. */
	   
	   /****************************************************************************
	   * Create a 1-dimensional gaussian smoothing kernel.
	   ****************************************************************************/
	   while(1)
	   {
	      windowsize = 42; // make_gaussian_kernel(SIGMA, gaussian_kernel, &windowsize);
	      center = windowsize / 2;	   
	      kernel_center = center;
	      G_Out1.write(gaussian_kernel);
	      G_Out2.write(gaussian_kernel);
	      C_Out1.write(kernel_center);
	      C_Out2.write(kernel_center);
	   }
	}
	
	SC_CTOR(Gaussian_Kernel)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(BlurX)
{
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_in<FKERNAL> KernelIn;
	sc_fifo_in<int> CenterIn;
	sc_fifo_out<FIMAGE> TempimOut;
	
	IMAGE image;
	FKERNAL kernel;
	int center;
	FIMAGE tempim;
	
	void main(void)
	{
	   while(1)
       	   {
	      ImgIn.read(image);
	      KernelIn.read(kernel);
	      CenterIn.read(center);
	      tempim = image; //blur_x(ROWS, COLS);
	      wait(1710, SC_MS);
	      TempimOut.write(tempim);	
	   }
	}
	
	SC_CTOR(BlurX)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(BlurY)
{	
	sc_fifo_in<FKERNAL> KernelIn;
	sc_fifo_in<int> CenterIn;
	sc_fifo_in<FIMAGE> TempimIn;
	sc_fifo_out<SIMAGE> SmoothedimOut;
	
	FKERNAL kernel;
	int center;
	FIMAGE tempim;
	SIMAGE smoothedim;

	void main(void)
	{
	   while(1)
	   {
	      TempimIn.read(tempim);
	      KernelIn.read(kernel);
	      CenterIn.read(center);
	      smoothedim = tempim; // blur_y(ROWS, COLS);
	      wait(1820, SC_MS);
	      SmoothedimOut.write(smoothedim);
	   }
	}
	
	SC_CTOR(BlurY)
	{
	   SC_THREAD(main);
	}
};

SC_MODULE(Gaussian_Smooth)
{	
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_out<SIMAGE> SmoothedimOut;
	
	sc_fifo<IMAGE> img_q;
	sc_fifo<FKERNAL> knl_q1,knl_q2;
	sc_fifo<FIMAGE> tmp_q;
	sc_fifo<int> int_q1,int_q2;
	
	Receive_Image receiver;
	Gaussian_Kernel g_kernel;
	BlurX blurx;
	BlurY blury;
	
	SC_CTOR(Gaussian_Smooth) 
	:img_q("img_q",1)
	,knl_q1("knl_q1",1)
	,knl_q2("knl_q2", 1)
	,tmp_q("tmp_q",1)
	,int_q1("int_q1",1)
	,int_q2("int_q2",1)
	,receiver("receiver")
	,g_kernel("g_kernel")
	,blurx("blurx")
	,blury("blury")
 	{
           receiver.ImgIn.bind(ImgIn);
           receiver.ImgOut.bind(img_q);

           g_kernel.C_Out1.bind(int_q1);
           g_kernel.C_Out2.bind(int_q2);
           g_kernel.G_Out1.bind(knl_q1);
           g_kernel.G_Out2.bind(knl_q2);

           blurx.ImgIn.bind(img_q);
           blurx.CenterIn.bind(int_q1);
           blurx.KernelIn.bind(knl_q1);
           blurx.TempimOut.bind(tmp_q);

           blury.CenterIn.bind(int_q2);
           blury.KernelIn.bind(knl_q2);
           blury.TempimIn.bind(tmp_q);
           blury.SmoothedimOut.bind(SmoothedimOut);
	}
};

SC_MODULE(Derivative_X_Y)
{
	sc_fifo_in<SIMAGE>	SmoothedimIn;
	sc_fifo_out<SIMAGE> XOut1, XOut2, YOut1, YOut2;
	SIMAGE smoothedim, delta_x, delta_y;

	void main(void)
	{
	   while(1)
	   {
	      SmoothedimIn.read(smoothedim);
	      delta_x = delta_y = smoothedim; // derivative_x_y(ROWS, COLS);
	      wait(480, SC_MS);
	      XOut1.write(delta_x);
	      XOut2.write(delta_x);
	      YOut1.write(delta_y);
	      YOut2.write(delta_y);
	   }
	}
	
	SC_CTOR(Derivative_X_Y)
	{
           SC_THREAD(main);
    	}
};

SC_MODULE(Magnitude_X_Y)
{
	sc_fifo_in<SIMAGE> XIn, YIn;
	sc_fifo_out<SIMAGE>	MagnitudeOut1, MagnitudeOut2;
	SIMAGE delta_x, delta_y, magnitude;

	void main(void)
	{
	   while(1)
	   {
	      XIn.read(delta_x);
	      YIn.read(delta_y);
	      magnitude = delta_x + delta_y; //magnitude_x_y(ROWS, COLS);
	      wait(1030, SC_MS);
	      MagnitudeOut1.write(magnitude);
	      MagnitudeOut2.write(magnitude);
	   }
	}
	
	SC_CTOR(Magnitude_X_Y)
	{
           SC_THREAD(main);
	}
};

SC_MODULE(Non_Max_Supp)
{
	sc_fifo_in<SIMAGE> GradxIn, GradyIn, MagIn;
	sc_fifo_out<IMAGE> NmsOut;
	SIMAGE gradx, grady, mag;
	IMAGE nms;

	void main(void)
	{
	   IMAGE result;

	   while(1)
	   {
	      GradxIn.read(gradx);
	      GradyIn.read(grady);
	      MagIn.read(mag);    
	      result = gradx + grady + mag; // non_max_supp(ROWS, COLS, result);
	      wait(830, SC_MS);
	      nms = result;
	      NmsOut.write(nms);
	   }
	}
	
	SC_CTOR(Non_Max_Supp)
	{
           SC_THREAD(main);
	}
};

SC_MODULE(Apply_Hysteresis)
{
	sc_fifo_in<SIMAGE> MagIn;
	sc_fifo_in<IMAGE> NmsIn;
	sc_fifo_out<IMAGE> ImgOut;
	SIMAGE mag;
	IMAGE nms, EdgeImage;
	
	void main(void)
	{
	   while(1)
	   {
	      MagIn.read(mag);
	      NmsIn.read(nms);
	      EdgeImage = mag + nms; // apply_hysteresis(ROWS, COLS, TLOW, THIGH, EdgeImage);
	      wait(670, SC_MS);
	      ImgOut.write(EdgeImage);
	   }
	}
	
	SC_CTOR(Apply_Hysteresis)
	{
           SC_THREAD(main);
	}
};

SC_MODULE(DUT)
{
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_out<IMAGE> ImgOut;
	sc_fifo<SIMAGE> q1, q2, q3, q4, q5, q6, q7;
	sc_fifo<IMAGE> nms;

	Gaussian_Smooth gaussian_smooth;
	Derivative_X_Y derivative_x_y;
	Magnitude_X_Y magnitude_x_y;
	Non_Max_Supp non_max_supp;
	Apply_Hysteresis apply_hysteresis;

	SC_CTOR(DUT)
	:q1("q1",1)
	,q2("q2",1)
	,q3("q3",1)
	,q4("q4",1)
	,q5("q5",1)
	,q6("q6",1)
	,q7("q7",1)
	,nms("nms",1)
	,gaussian_smooth("gaussian_smooth")
	,derivative_x_y("derivative_x_y")
	,magnitude_x_y("magnitude_x_y")
	,non_max_supp("non_max_supp")
	,apply_hysteresis("apply_hysteresis")
	{
	   gaussian_smooth.ImgIn.bind(ImgIn);
	   gaussian_smooth.SmoothedimOut.bind(q1);
		        
           derivative_x_y.SmoothedimIn.bind(q1);
           derivative_x_y.XOut1.bind(q2);
           derivative_x_y.XOut2.bind(q3);
           derivative_x_y.YOut1.bind(q4);
           derivative_x_y.YOut2.bind(q5);

           magnitude_x_y.XIn.bind(q2);
           magnitude_x_y.YIn.bind(q4);
           magnitude_x_y.MagnitudeOut1.bind(q6);
           magnitude_x_y.MagnitudeOut2.bind(q7);

           non_max_supp.GradxIn.bind(q3);
           non_max_supp.GradyIn.bind(q5);
           non_max_supp.MagIn.bind(q6);
           non_max_supp.NmsOut.bind(nms);

           apply_hysteresis.MagIn.bind(q7);
           apply_hysteresis.NmsIn.bind(nms);
           apply_hysteresis.ImgOut.bind(ImgOut);
	}
};

SC_MODULE(Platform)
{
	sc_fifo_in<IMAGE> ImgIn;
	sc_fifo_out<IMAGE> ImgOut;
	sc_fifo<IMAGE> q1;
	sc_fifo<IMAGE> q2;

	DataIn din;
	DUT canny;
	DataOut dout;

	void before_end_of_elaboration(){
	   din.ImgIn.bind(ImgIn);
	   din.ImgOut.bind(q1);
	   canny.ImgIn.bind(q1);
	   canny.ImgOut.bind(q2);
	   dout.ImgIn.bind(q2);
	   dout.ImgOut.bind(ImgOut);
	}

	SC_CTOR(Platform)
	:q1("q1",1)
	,q2("q2",1)
	,din("din")
	,canny("canny")
	,dout("dout")
	{}
};

SC_MODULE(Top)
{
	sc_fifo<IMAGE> q1;
	sc_fifo<IMAGE> q2;
	sc_fifo<sc_time> qt;
	Stimulus stimulus;
	Platform platform;
	Monitor monitor;

	void before_end_of_elaboration(){
	   stimulus.ImgOut.bind(q1);
	   stimulus.StartTime.bind(qt);
	   platform.ImgIn.bind(q1);
	   platform.ImgOut.bind(q2);
	   monitor.ImgIn.bind(q2);
	   monitor.StartTime.bind(qt);
	}

	SC_CTOR(Top)
	:q1("q1",1)
	,q2("q2",1)
	,qt("qt",30)
	,stimulus("stimulus")
	,platform("platform")
	,monitor("monitor")
	{}
};

Top top("top");

int sc_main(int argc, char* argv[])
{
	sc_time ExpectedTotalTime(59320, SC_MS);

	sc_start();
	assert(sc_time_stamp() == ExpectedTotalTime);
	return 0;
}

// EOF
