// parallel_threads2.cc: regression test for parallel execution
// RD, 04/12/17

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS 10
#define WORKLOAD   10000000

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Worker1)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(1, SC_NS);
	Count++;
	w = Work(WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(WORKLOAD);
	}
	assert(Count == ITERATIONS+1);
        sc_time expected_time(1+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker1)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker1);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker2)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(2, SC_NS);
	Count++;
	w = Work(WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(WORKLOAD);
	}
	assert(Count == ITERATIONS+1);
        sc_time expected_time(2+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker2)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker2);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    Worker1 w1, w3;	// odd times
    Worker2 w2, w4;	// even times

    SC_CTOR(TOP)
    : w1("w1"), w3("w3"), w2("w2"), w4("w4")
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("parallel_threads2 starts\n");
    sc_start();
    assert(top.w1.Count == ITERATIONS+1);
    assert(top.w2.Count == ITERATIONS+1);
    assert(top.w3.Count == ITERATIONS+1);
    assert(top.w4.Count == ITERATIONS+1);
    sc_time expected_time(2+10*ITERATIONS, SC_NS);
    assert(sc_time_stamp() == expected_time);
    printf("parallel_threads2 done.\n");
    printf("[potential speedup over SEQ: OOO 4x, NPD 4x, SYN 2x]\n");
    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
