// parallel_threads3.cc: regression test for parallel execution
// RD, 04/12/17

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS 10
#define WORKLOAD   10000000

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Worker1)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(1, SC_NS);
	Count++;
	w = Work(WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(WORKLOAD);
	}
	assert(Count == ITERATIONS+1);
        sc_time expected_time(1+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker1)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker1);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker2)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(2, SC_NS);
	Count++;
	w = Work(WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(WORKLOAD);
	}
	assert(Count == ITERATIONS+1);
        sc_time expected_time(2+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker2)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker2);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker3)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(3, SC_NS);
	Count++;
	w = Work(WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(WORKLOAD);
	}
	assert(Count == ITERATIONS+1);
        sc_time expected_time(3+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker3)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker3);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker4)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(4, SC_NS);
	Count++;
	w = Work(WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(WORKLOAD);
	}
	assert(Count == ITERATIONS+1);
        sc_time expected_time(4+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker4)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker4);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    Worker1 w1a, w1b;	// working at time 1, 11, 21, etc.
    Worker2 w2a, w2b;	// working at time 2, 12, 22, etc.
    Worker3 w3a, w3b;	// working at time 3, 13, 23, etc.
    Worker4 w4a, w4b;	// working at time 4, 14, 24, etc.

    SC_CTOR(TOP)
    :   w1a("w1a"), w1b("w1b"),
        w2a("w2a"), w2b("w2b"),
        w3a("w3a"), w3b("w3b"),
        w4a("w4a"), w4b("w4b")
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("parallel_threads3 starts\n");
    sc_start();
    assert(top.w1a.Count == ITERATIONS+1);
    assert(top.w2a.Count == ITERATIONS+1);
    assert(top.w3a.Count == ITERATIONS+1);
    assert(top.w4a.Count == ITERATIONS+1);
    assert(top.w1b.Count == ITERATIONS+1);
    assert(top.w2b.Count == ITERATIONS+1);
    assert(top.w3b.Count == ITERATIONS+1);
    assert(top.w4b.Count == ITERATIONS+1);
    sc_time expected_time(4+10*ITERATIONS, SC_NS);
    assert(sc_time_stamp() == expected_time);
    printf("parallel_threads3 done.\n");
    printf("[potential speedup over SEQ: OOO 8x, NPD 8x, SYN 2x]\n");
    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
