// rendezvous.cc (WAS: particle_simulation.cpp)
// 04/29/19, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define NSTEPS 100

int Lines = 0;

class i_transceiver: virtual public sc_interface
{
    public:
	virtual void read(int &) = 0;
	virtual void write(const int &) = 0;
};

class c_double_handshake_channel:
	public sc_channel, public i_transceiver
{
    public:
	c_double_handshake_channel(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_double_handshake_channel()
	{ }

	void read(int &Y)
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    Y = tmpd;

	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	}

	void write(const int &X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}
    private:
	sc_event req;
	sc_event ack;
	bool v, w;
	int tmpd;
};

SC_MODULE(Tile)
{
    sc_port<i_transceiver> North;
    sc_port<i_transceiver> East;
    sc_port<i_transceiver> South;
    sc_port<i_transceiver> West;

    void main(void);
    void talk2neighbors(void);

    Tile(sc_module_name name_, int Row_ = 1, int Col_ = 1):
	sc_module(name_),
	Row(Row_), Col(Col_),
	Count((Row-1)*2+Col)
    {
	SC_HAS_PROCESS(Tile);
	SC_THREAD(main);
    }

    private:
	const int Row, Col;
    public:
	int Count;
};

void Tile::main(void)
{
	printf("  Tile[r%d,c%d] starts, Count=%d.\n", Row, Col, Count);
	assert(Count == (Row-1)*2+Col);

	for(int step = 0; step < NSTEPS; step++)
	{
	    talk2neighbors();
	    if(++Lines % 4 == 0)
		printf("\t---------------------------------------------\n");
	}

	wait(1,SC_NS);
	printf("  Tile[r%d,c%d] ends, Count=%d.\n", Row, Col, Count);
	assert(Count == NSTEPS*2 + (Row-1)*2+Col);
}

void Tile::talk2neighbors(void)
{
	if (Row % 2 == 0)	//even row
	{
	    int Value;
	    South->read(Value);
	    Value++;
	    South->write(Value);
	    North->write(Count);
	    North->read(Count);
	}
	else
	{
	    int Value;
	    North->write(Count);
	    North->read(Count);
	    South->read(Value);
	    Value++;
	    South->write(Value);
	}

	if (Col % 2 == 0)	// even column
	{
	    int Value;
	    West->read(Value);
	    Value++;
	    West->write(Value);
	    East->write(Count);
	    East->read(Count);
	}
	else
	{
	    int Value;
	    East->write(Count);
	    East->read(Count);
	    West->read(Value);
	    Value++;
	    West->write(Value);
	}
	printf("%sTile[r%d,c%d] Count=%d\n",
		"\t\t\t\t"+(4-(Row-1)*2-Col),
		Row, Col, Count);
}

SC_MODULE(TOP)
{
    //straight links
    c_double_handshake_channel   c23;
    c_double_handshake_channel   c32;
    c_double_handshake_channel   c43;
    c_double_handshake_channel   c34;
    //torus links
    c_double_handshake_channel   c_h1;
    c_double_handshake_channel   c_h2;
    c_double_handshake_channel   c_v1;
    c_double_handshake_channel   c_v2;

    Tile	t1;
    Tile	t2;
    Tile	t3;
    Tile	t4;

    SC_CTOR(TOP)
    : c23("c23"), c32("c32"), c43("c43"), c34("c34"),
	c_h1("c_h1"), c_h2("c_h2"), c_v1("c_v1"), c_v2("c_v2"),
	t1("t1",1, 1), t2("t2",1, 2), t3("t3", 2, 1), t4("t4",2, 2)
    {
	// port binding
	t1(c_v1, c23, c32, c_h1);
	t2(c_v2, c_h1, c34, c23);
	t3(c32, c43, c_v1, c_h2);
	t4(c34, c_h2, c_v2, c43);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("rendezvous starts\n");
    sc_start();
    assert(top.t1.Count == NSTEPS*2 + 1);
    assert(top.t2.Count == NSTEPS*2 + 2);
    assert(top.t3.Count == NSTEPS*2 + 3);
    assert(top.t4.Count == NSTEPS*2 + 4);
    assert(Lines == NSTEPS*4);
    printf("rendezvous done.\n");
    return 0;
}

// EOF
