// run_ahead2b.cc: regression test for parallel execution
// (same as run_ahead2.cc, but with multiple sc_start() calls)
// RD, 08/29/19

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS 10
#define LIGHT_WORKLOAD      50000
#define HEAVY_WORKLOAD   50000000

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Worker1)	// works at 1 ns, 11 ns, 21 ns, ...
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(1, SC_NS);
	cout << sc_time_stamp() << ": worker1 starts light work.\n";
	Count++;
	w = Work(LIGHT_WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(LIGHT_WORKLOAD);
	}
	cout << sc_time_stamp() << ": worker1 light work done.\n";
	assert(Count == ITERATIONS+1);
	sc_time expected_time(1+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);

	cout << sc_time_stamp() << ": worker1 stopping simulation...\n";
	sc_stop();
    }

    SC_CTOR(Worker1)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker1);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker2)	// works at 2 ns, 12 ns, 22 ns, ...
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(2, SC_NS);
	cout << sc_time_stamp() << ": worker2 starts heavy work.\n";
	Count++;
	w = Work(HEAVY_WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(HEAVY_WORKLOAD);
	}
	cout << sc_time_stamp() << ": worker2 heavy work done (NOT EXPECTED!).\n";
	assert(false);	// worker 2 won't finish!
    }

    SC_CTOR(Worker2)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker2);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    Worker1 w1;
    Worker2 w2;

    SC_CTOR(TOP)
    : w1("w1"), w2("w2")
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    sc_time expected_time1(1*30, SC_NS);
    sc_time expected_time2(2*30, SC_NS);
    sc_time expected_time3(3*30, SC_NS);
    sc_time expected_time4(1+10*ITERATIONS, SC_NS);

    cout << sc_time_stamp() << ": run_ahead2b starts\n";
    sc_start(30, SC_NS /* SC_RUN_TO_TIME */);
    cout << sc_time_stamp() << ": run_ahead2b 1 quarter done.\n";
    assert(sc_time_stamp() == expected_time1);
    assert(top.w1.Count == 3);
    assert(top.w2.Count == 3);
    sc_start(30, SC_NS /* SC_RUN_TO_TIME */);
    cout << sc_time_stamp() << ": run_ahead2b 2 quarters done.\n";
    assert(sc_time_stamp() == expected_time2);
    assert(top.w1.Count == 6);
    assert(top.w2.Count == 6);
    sc_start(30, SC_NS /* SC_RUN_TO_TIME */);
    cout << sc_time_stamp() << ": run_ahead2b 3 quarters done.\n";
    assert(sc_time_stamp() == expected_time3);
    assert(top.w1.Count == 9);
    assert(top.w2.Count == 9);
    sc_start(30, SC_NS /* SC_RUN_TO_TIME */);
    cout << sc_time_stamp() << ": run_ahead2b quits\n";
    assert(sc_time_stamp() == expected_time4);
    assert(top.w1.Count == ITERATIONS+1);
    assert(top.w2.Count == ITERATIONS);		// worker 2 won't finish!
    return 0;
}

// EOF
