#include <iostream>

#include "risc.h"

int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);


  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;


      sg.add_function(*iter_thread, *iter_thread);

      //AstDOTGeneration dotgen;
      //dotgen.generate((*iter_thread)->get_ast_node(), (*iter_thread)->get_name());
    }

    std::cout << std::endl;
  }

  sg.clean_graph();

  risc::sg::DataConflictTable dct(sg, d.path_instance_mapper_, true);

  risc::sg::CombinedDataConflictTable cdct(dct, d.path_instance_mapper_);

  if(d.perform_static_analysis()) {
    sg.print_graph("combined_data_conflict_table_static.dot");
    sg.print_graph_read_write_access("combined_data_conflict_table_read_write_static.dot");
    cdct.print_conflict_table("combined_data_conflict_table_static.html");

  } else {

    sg.print_graph("combined_data_conflict_table_dynamic.dot");
    sg.print_graph_read_write_access("combined_data_conflict_table_read_write_dynamic.dot");
    cdct.print_conflict_table("combined_data_conflict_table_dynamic.html");
  }

  cdct.print_lookup_table("combined_data_conflict_table_lookup_table.html");

  std::cout << "Printing combined data conflict table:" << std::endl;
  cdct.print_conflict_table();

  std::cout << "Creating event conflict table" << std::endl;
  risc::sg::EventConflictTable ect(sg, d.path_instance_mapper_, true);

  std::cout << "Creating time advance table" << std::endl;
  risc::sg::TimeAdvanceTable time_advance_table(sg);

  risc::tools::ParseStatus status;
  std::vector< std::vector<int> > psg_seg_id_array(1);
  std::vector<std::string> input_psg_file_prefix;
  std::vector< std::vector<int> > channel_seg_ids(1);
  std::vector<std::string> channel_seg_id_array_names_;
  std::cout << "Instrumenting file" << std::endl;
  risc::inst::instrumentor(d, sg, &dct, &ect, &time_advance_table, &cdct, NULL, NULL, NULL, &status,"",psg_seg_id_array,input_psg_file_prefix,
    channel_seg_ids, channel_seg_id_array_names_);

  std::cout << "Unparse" << std::endl;
  risc::inst::risc_unparse(d, &status, "", NULL, NULL);

  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
