#include <iostream>

#include "risc.h"

int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);


  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  typedef std::pair<int, risc::ThreadVectorIter> IdThreadPair;
  std::list<IdThreadPair> starting_ids;
  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;


      sg.add_function(*iter_thread, *iter_thread);

      starting_ids.push_back(IdThreadPair((*iter_thread)->starting_id_, iter_thread));
      //AstDOTGeneration dotgen;
      //dotgen.generate((*iter_thread)->get_ast_node(), (*iter_thread)->get_name());
    }

    std::cout << std::endl;
  }

  sg.clean_graph();

  // generating consecutive numbers
#if 1
  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;
  std::pair<VertexIterator, VertexIterator> vp;
  int i = 0;
  for(vp = boost::vertices(sg.graph_);
      vp.first != vp.second;
      ++vp.first, i++) {

    bool found = false;
    std::list<IdThreadPair>::iterator iter;
    for(iter  = starting_ids.begin();
        iter != starting_ids.end();
        iter++) {

      if((*iter).first == sg.graph_[*vp.first].id_) {
        found = true;
        break;
      }
    }

    sg.graph_[*vp.first].id_ = i;
    if(found)  {
      IdThreadPair temp = *iter;
      risc::ThreadVectorIter temp_iter= temp.second;
      (*temp_iter)->starting_id_ = i;
      starting_ids.erase(iter);
    }
  }
#endif

  risc::sg::DataConflictTable dct(sg, d.path_instance_mapper_, true);


  std::cout << "Print combined data conflict table:" << std::endl;

  std::cout << "Creating event conflict table" << std::endl;
  risc::sg::EventConflictTable ect(sg, d.path_instance_mapper_, true);
  ect.print_conflict_table("prediction_event_conflict_table.html");
  ect.print_lookup_table("prediction_event_lookup_table.html");


  std::cout << "Creating prediction event notification table" << std::endl;
  risc::sg::TimeAdvanceTable time_advance_table(sg);

  risc::sg::PredictionEventNotificationTable pent(ect, d.path_instance_mapper_);
  if(d.perform_static_analysis()) {
    sg.print_graph("prediction_event_notification_table_static.dot");
    sg.print_graph_read_write_access("prediction_event_notification_table_read_write_static.dot");
    pent.print_notification_table("prediction_event_notification_table_static.html");

  } else {

    sg.print_graph("prediction_event_notification_table_dynamic.dot");
    sg.print_graph_read_write_access("prediction_event_notification_table_read_write_dynamic.dot");
    pent.print_notification_table("prediction_event_notification_table_dynamic.html");
  }
  pent.print_instance_id_to_index_lookup_table("prediction_event_notification_lookup_table.html");

  std::cout << "Printing prediction event notification table:" << std::endl;
  pent.print_notification_table();

  std::cout << "Printing prediction event notification lookup table:" << std::endl;
  pent.print_instance_id_to_index_lookup_table();

  std::cout << "Printing event notification table" << std::endl;
  pent.ect_.print_conflict_table("event_notification_table_for_prediction.html");

  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
